@@ -5,7 +5,7 @@ ${CXX:-g++} -fprofile-arcs -ftest-coverage -fPIC -O0 example.cpp -o program
 ./program
 
 #BEGIN gcovr
-gcovr -r . --branches
+gcovr --branches
 #END gcovr
 
 rm -f program *.gc*
